package liv.cache.ahead;

/**
 * Interface which defines the basic behaviour of items which can be created
 * ahead and cached (pooled). It simply defines a way for subclasses to
 * "auto-destruct" and notify the pool/cached whether they are in a clean state
 * (and as such they can be returned back to the pool).
 *
 * @author Liviu Tudor http://about.me/liviutudor
 */
public interface CacheableAhead {
    /**
     * Resets the object to the "initial" state. That is the state the object
     * was in right after it was created (via <code>new</code>).
     */
    void activate();

    /**
     * Prepares the object to be returned to the pool. Needs to free up any
     * resources at any point. Similar to {@link #destroy()} but potentially
     * less "destructive".
     */
    void passivate();

    /**
     * "Destroys" the object. That is, frees up any resources that the object is
     * taking and prepares it to be released into the great void of garbage
     * collection :D
     */
    void destroy();

    /**
     * Used by the caches/pools to verify whether this object is valid to be
     * re-pooled or it needs to be destroyed and chucked away.
     *
     * @return true if the object is valid and can be returned to the pool
     *         (passivated first) or false if this object needs to be thrown
     *         away (after it's been destroyed)
     */
    boolean isValid();
}
