package liv.cache.ahead;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.lang.reflect.Field;
import java.util.Set;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Unit test for {@link CreateInstanceAheadCache}.
 *
 * @author Liviu Tudor http://about.me/liviutudor
 */
public class CreateInstanceAheadCacheTest {
    static Field      fieldBackgroundTask;
    static Field      fieldKeys;
    static final int  DEF_MIN_IDLE        = 5;
    static final long DEF_WAKEUP_INTERVAL = 100L; // 100ms

    /**
     * Used internally to make certain private fields of the
     * {@link CreateInstanceAheadCache} class available for reading/writing.
     *
     * @throws Exception
     *             if any reflection exceptions occur
     */
    static void prepareClass() throws Exception {
        fieldBackgroundTask = CreateInstanceAheadCache.class.getDeclaredField("backgroundTask");
        fieldBackgroundTask.setAccessible(true);

        fieldKeys = CreateInstanceAheadCache.class.getDeclaredField("keys");
        fieldKeys.setAccessible(true);
    }

    static ScheduledExecutorService getBackgroundTask(CreateInstanceAheadCache<?> obj) throws Exception {
        return (ScheduledExecutorService) fieldBackgroundTask.get(obj);
    }

    @SuppressWarnings("unchecked")
    static <T extends CacheableAhead> Set<Class<T>> getKeys(CreateInstanceAheadCache<T> obj) throws Exception {
        return (Set<Class<T>>) fieldKeys.get(obj);
    }

    @BeforeClass
    public static void setUp() throws Exception {
        prepareClass();
    }

    @Test
    public void testCreateWrongParameters() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        CreateInstanceAheadCache<DummyCacheableAhead> cache = null;
        boolean error = false;

        // wrong min idle
        error = false;
        try {
            cache = new CreateInstanceAheadCache<DummyCacheableAhead>(factory, 0, 100L);
        } catch (IllegalArgumentException e) {
            error = true;
        }
        assertTrue(error);

        error = false;
        try {
            cache = new CreateInstanceAheadCache<DummyCacheableAhead>(factory, -5, 100L);
        } catch (IllegalArgumentException e) {
            error = true;
        }
        assertTrue(error);

        // wrong interval
        error = false;
        try {
            cache = new CreateInstanceAheadCache<DummyCacheableAhead>(factory, DEF_MIN_IDLE, 0L);
        } catch (IllegalArgumentException e) {
            error = true;
        }
        assertTrue(error);

        error = false;
        try {
            cache = new CreateInstanceAheadCache<DummyCacheableAhead>(factory, DEF_MIN_IDLE, -9L);
        } catch (IllegalArgumentException e) {
            error = true;
        }
        assertTrue(error);

        // wrong both
        error = false;
        try {
            cache = new CreateInstanceAheadCache<DummyCacheableAhead>(factory, 0, 0L);
        } catch (IllegalArgumentException e) {
            error = true;
        }
        assertTrue(error);

        assertNull(cache);
    }

    @Test
    public void checkTestObjectInit() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        // schedule a really long interval and ensure all counters are null
        DummyCreateInstanceAheadCache cache = new DummyCreateInstanceAheadCache(factory, DEF_MIN_IDLE,
                TimeUnit.DAYS.toMillis(1L));
        assertEquals(cache.getCloseCalls(), 0);
        assertEquals(cache.getTotalBorrowObjectCalls(), 0);
        assertEquals(cache.getTotalClearCalls(), 0);
        assertEquals(cache.getTotalPrepopulateCall(), 0);
        assertEquals(cache.getClearAllCalls(), 0);

        cache.close();
    }

    @Test
    public void testClose() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        // schedule a really long interval so timer doesn't kick in
        DummyCreateInstanceAheadCache cache = new DummyCreateInstanceAheadCache(factory, DEF_MIN_IDLE,
                TimeUnit.DAYS.toMillis(1L));
        assertEquals(cache.getCloseCalls(), 0);

        cache.close();
        assertEquals(cache.getCloseCalls(), 1);
        ScheduledExecutorService backgroundTask = getBackgroundTask(cache);
        // give it time to complete the above call
        TimeUnit.MILLISECONDS.sleep(100L);
        assertTrue(backgroundTask.isTerminated());
        assertTrue(backgroundTask.isShutdown());
    }

    @Test
    public void testBorrowPrepopulateClear() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCreateInstanceAheadCache cache = new DummyCreateInstanceAheadCache(factory, DEF_MIN_IDLE,
                DEF_WAKEUP_INTERVAL);
        // borrow then return
        DummyCacheableAhead obj = cache.borrowObject(DummyCacheableAhead.class);
        assertNotNull(obj);
        assertEquals(cache.getBorrowObjectCalls(DummyCacheableAhead.class), 1);
        assertEquals(cache.getNumActive(), 1);
        assertEquals(cache.getNumIdle(), 0);
        // return
        cache.returnObject(DummyCacheableAhead.class, obj);
        // give enough time for the prepopulate to kick in
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        // test prepopulate
        assertEquals(cache.getPrepopulateCalls(DummyCacheableAhead.class), TimeUnit.SECONDS.toMillis(1)
                / DEF_WAKEUP_INTERVAL);
        assertEquals(cache.getNumIdle(), DEF_MIN_IDLE);
        // now test clear
        cache.clear(DummyCacheableAhead.class);
        assertEquals(cache.getClearCalls(DummyCacheableAhead.class), 1);
        assertEquals(cache.getNumIdle(), 0);
        // give it enough time to ensure that prepopulate doesn't wipe that out
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        // no more keys in set so this shouldn't count!!
        assertEquals(cache.getPrepopulateCalls(DummyCacheableAhead.class), TimeUnit.SECONDS.toMillis(1)
                / DEF_WAKEUP_INTERVAL);
        assertEquals(cache.getNumIdle(), 0);

        // now ensure the second clear gets called too, even though key is no
        // longer in our set of keys
        cache.clear(DummyCacheableAhead.class);
        assertEquals(cache.getClearCalls(DummyCacheableAhead.class), 2);

        // now clear all and ensure the hook gets called
        cache.clear();
        assertEquals(cache.getClearAllCalls(), 1);

        cache.close();
    }

    @Test
    public void testKeys() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCreateInstanceAheadCache cache = new DummyCreateInstanceAheadCache(factory, DEF_MIN_IDLE,
                DEF_WAKEUP_INTERVAL);
        Set<Class<DummyCacheableAhead>> keys = getKeys(cache);
        assertNotNull(keys);
        assertEquals(keys.size(), 0);
        // ensure that even after the task kicks in we still have 0 keys
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 0);

        /*
         * now borrow one object and don't put it back, and this should see the
         * key size go up
         */
        DummyCacheableAhead obj = cache.borrowObject(DummyCacheableAhead.class);
        assertEquals(keys.size(), 1);
        assertTrue(keys.contains(DummyCacheableAhead.class));
        // ensure that's still the case after prepopulate
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 1);
        assertTrue(keys.contains(DummyCacheableAhead.class));

        // now delete the key
        cache.clear(DummyCacheableAhead.class);
        System.out.println("In test: " + keys);
        // just in case the other thread is running still, give it some time
        // TimeUnit.MILLISECONDS.sleep(DEF_WAKEUP_INTERVAL );
        assertEquals(keys.size(), 0);
        // wait for prepopulate to kick in and ensure that's still the case
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 0);

        /*
         * borrow an object, put it back and ensure everything is still
         * happening the same
         */
        obj = cache.borrowObject(DummyCacheableAhead.class);
        cache.returnObject(DummyCacheableAhead.class, obj);
        assertEquals(keys.size(), 1);
        assertTrue(keys.contains(DummyCacheableAhead.class));
        // ensure that's still the case after prepopulate
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 1);
        assertTrue(keys.contains(DummyCacheableAhead.class));

        // now delete the key
        cache.clear(DummyCacheableAhead.class);
        assertEquals(keys.size(), 0);
        // wait for prepopulate to kick in and ensure that's still the case
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 0);

        /*
         * borrow an object, and call just clear and ensure everything is the
         * same
         */
        obj = cache.borrowObject(DummyCacheableAhead.class);
        assertEquals(keys.size(), 1);
        assertTrue(keys.contains(DummyCacheableAhead.class));
        // ensure that's still the case after prepopulate
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 1);
        assertTrue(keys.contains(DummyCacheableAhead.class));

        // now delete the key
        cache.clear();
        assertEquals(keys.size(), 0);
        // wait for prepopulate to kick in and ensure that's still the case
        TimeUnit.SECONDS.sleep(1); // wait 1 second
        assertEquals(keys.size(), 0);

        cache.close();
    }
}
