package liv.cache.ahead;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.Test;

/**
 * Unit test for {@link CreateInstanceFactory}.
 *
 * @author Liviu Tudor http://about.me/liviutudor
 */
public class CreateInstanceFactoryTest {
    @Test
    public void testMakeObject() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCacheableAhead dummy = factory.makeObject(DummyCacheableAhead.class);
        assertEquals(dummy.getActivateCalls(), 0);
        assertEquals(dummy.getDestroyCalls(), 0);
        assertEquals(dummy.getIsValidCalls(), 0);
        assertEquals(dummy.getPassivateCalls(), 0);
        assertTrue(dummy.isValid());
    }

    @Test
    public void testValidateObject() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCacheableAhead dummyValid = new DummyCacheableAhead(true);
        DummyCacheableAhead dummyInvalid = new DummyCacheableAhead(false);
        boolean validResponse = factory.validateObject(DummyCacheableAhead.class, dummyValid);
        boolean invalidResponse = factory.validateObject(DummyCacheableAhead.class, dummyInvalid);
        assertTrue(validResponse);
        assertEquals(dummyValid.getActivateCalls(), 0);
        assertEquals(dummyValid.getDestroyCalls(), 0);
        assertEquals(dummyValid.getIsValidCalls(), 1);
        assertEquals(dummyValid.getPassivateCalls(), 0);

        assertFalse(invalidResponse);
        assertEquals(dummyInvalid.getActivateCalls(), 0);
        assertEquals(dummyInvalid.getDestroyCalls(), 0);
        assertEquals(dummyInvalid.getIsValidCalls(), 1);
        assertEquals(dummyInvalid.getPassivateCalls(), 0);
    }

    @Test
    public void testDestroyObject() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCacheableAhead dummyValid = new DummyCacheableAhead(true);
        DummyCacheableAhead dummyInvalid = new DummyCacheableAhead(false);

        factory.destroyObject(DummyCacheableAhead.class, dummyValid);
        assertEquals(dummyValid.getActivateCalls(), 0);
        assertEquals(dummyValid.getDestroyCalls(), 1);
        assertEquals(dummyValid.getIsValidCalls(), 0);
        assertEquals(dummyValid.getPassivateCalls(), 0);

        factory.destroyObject(DummyCacheableAhead.class, dummyInvalid);
        assertEquals(dummyInvalid.getActivateCalls(), 0);
        assertEquals(dummyInvalid.getDestroyCalls(), 1);
        assertEquals(dummyInvalid.getIsValidCalls(), 0);
        assertEquals(dummyInvalid.getPassivateCalls(), 0);
    }

    @Test
    public void testActivateObject() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCacheableAhead dummyValid = new DummyCacheableAhead(true);
        DummyCacheableAhead dummyInvalid = new DummyCacheableAhead(false);

        factory.activateObject(DummyCacheableAhead.class, dummyValid);
        assertEquals(dummyValid.getActivateCalls(), 1);
        assertEquals(dummyValid.getDestroyCalls(), 0);
        assertEquals(dummyValid.getIsValidCalls(), 0);
        assertEquals(dummyValid.getPassivateCalls(), 0);

        factory.activateObject(DummyCacheableAhead.class, dummyInvalid);
        assertEquals(dummyInvalid.getActivateCalls(), 1);
        assertEquals(dummyInvalid.getDestroyCalls(), 0);
        assertEquals(dummyInvalid.getIsValidCalls(), 0);
        assertEquals(dummyInvalid.getPassivateCalls(), 0);
    }

    @Test
    public void testPassivateObject() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        DummyCacheableAhead dummyValid = new DummyCacheableAhead(true);
        DummyCacheableAhead dummyInvalid = new DummyCacheableAhead(false);

        factory.passivateObject(DummyCacheableAhead.class, dummyValid);
        assertEquals(dummyValid.getActivateCalls(), 0);
        assertEquals(dummyValid.getDestroyCalls(), 0);
        assertEquals(dummyValid.getIsValidCalls(), 0);
        assertEquals(dummyValid.getPassivateCalls(), 1);

        factory.passivateObject(DummyCacheableAhead.class, dummyInvalid);
        assertEquals(dummyInvalid.getActivateCalls(), 0);
        assertEquals(dummyInvalid.getDestroyCalls(), 0);
        assertEquals(dummyInvalid.getIsValidCalls(), 0);
        assertEquals(dummyInvalid.getPassivateCalls(), 1);
    }

    @Test
    public void integrationTest() throws Exception {
        CreateInstanceFactory<DummyCacheableAhead> factory = new CreateInstanceFactory<DummyCacheableAhead>();
        int tries = 10;
        /*
         * 1.Create object (valid), passivate, activate then destroy.
         */
        DummyCacheableAhead dummyValid = factory.makeObject(DummyCacheableAhead.class);
        for (int i = 1; i <= tries; i++) {
            factory.passivateObject(DummyCacheableAhead.class, dummyValid);
            assertEquals(dummyValid.getPassivateCalls(), i);
            factory.activateObject(DummyCacheableAhead.class, dummyValid);
            assertEquals(dummyValid.getActivateCalls(), i);
            boolean validate = factory.validateObject(DummyCacheableAhead.class, dummyValid);
            assertTrue(validate);
            assertEquals(dummyValid.getIsValidCalls(), i);
            factory.destroyObject(DummyCacheableAhead.class, dummyValid);
            assertEquals(dummyValid.getDestroyCalls(), i);
        }

        /*
         * 2.Create object (invalid), passivate, activate then destroy.
         */
        DummyCacheableAhead dummyInvalid = factory.makeObject(DummyCacheableAhead.class);
        dummyInvalid.setValid(false);
        for (int i = 1; i <= tries; i++) {
            factory.passivateObject(DummyCacheableAhead.class, dummyInvalid);
            assertEquals(dummyInvalid.getPassivateCalls(), i);
            factory.activateObject(DummyCacheableAhead.class, dummyInvalid);
            assertEquals(dummyInvalid.getActivateCalls(), i);
            boolean validate = factory.validateObject(DummyCacheableAhead.class, dummyInvalid);
            assertFalse(validate);
            assertEquals(dummyInvalid.getIsValidCalls(), i);
            factory.destroyObject(DummyCacheableAhead.class, dummyInvalid);
            assertEquals(dummyInvalid.getDestroyCalls(), i);
        }
    }
}
