package liv.cache.ahead;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Implementation of a cache ahead class which counts number of calls to hook
 * methods.
 *
 * @author Liviu Tudor http://about.me/liviutudor
 */
public class DummyCreateInstanceAheadCache extends CreateInstanceAheadCache<DummyCacheableAhead> {
    private int                                            closeCalls;
    private int                                            clearAllCalls;
    private Map<Class<DummyCacheableAhead>, AtomicInteger> borrowObjectCalls;
    private Map<Class<DummyCacheableAhead>, AtomicInteger> clearCalls;
    private Map<Class<DummyCacheableAhead>, AtomicInteger> prepopulateCalls;

    public DummyCreateInstanceAheadCache(CreateInstanceFactory<DummyCacheableAhead> factory, int minIdle,
            long wakeUpInterval) {
        super(factory, minIdle, wakeUpInterval);
        clearAllCalls = closeCalls = 0;
        borrowObjectCalls = new HashMap<Class<DummyCacheableAhead>, AtomicInteger>();
        clearCalls = new HashMap<Class<DummyCacheableAhead>, AtomicInteger>();
        prepopulateCalls = new HashMap<Class<DummyCacheableAhead>, AtomicInteger>();
    }

    @Override
    protected void closeHook() {
        closeCalls++;
    }

    public int getCloseCalls() {
        return closeCalls;
    }

    @Override
    public void clearHook() {
        clearAllCalls++;
    }

    public int getClearAllCalls() {
        return clearAllCalls;
    }

    @Override
    protected void borrowObjectHook(Class<DummyCacheableAhead> key) {
        AtomicInteger newCount = new AtomicInteger();
        AtomicInteger count = borrowObjectCalls.get(key);
        if (count == null) {
            borrowObjectCalls.put(key, newCount);
            count = newCount;
        }
        count.incrementAndGet();
    }

    public int getBorrowObjectCalls(Class<DummyCacheableAhead> key) {
        AtomicInteger count = borrowObjectCalls.get(key);
        if (count == null) {
            return 0;
        }
        return count.intValue();
    }

    public int getTotalBorrowObjectCalls() {
        return borrowObjectCalls.size();
    }

    @Override
    protected void clearHook(Class<DummyCacheableAhead> key) {
        AtomicInteger newCount = new AtomicInteger();
        AtomicInteger count = clearCalls.get(key);
        if (count == null) {
            clearCalls.put(key, newCount);
            count = newCount;
        }
        count.incrementAndGet();
    }

    public int getClearCalls(Class<DummyCacheableAhead> key) {
        AtomicInteger count = clearCalls.get(key);
        if (count == null) {
            return 0;
        }
        return count.intValue();
    }

    public int getTotalClearCalls() {
        return clearCalls.size();
    }

    @Override
    protected void prepopulateHook(Class<DummyCacheableAhead> key) {
        AtomicInteger newCount = new AtomicInteger();
        AtomicInteger count = prepopulateCalls.get(key);
        if (count == null) {
            prepopulateCalls.put(key, newCount);
            count = newCount;
        }
        count.incrementAndGet();
    }

    public int getPrepopulateCalls(Class<DummyCacheableAhead> key) {
        AtomicInteger count = prepopulateCalls.get(key);
        if (count == null) {
            return 0;
        }
        return count.intValue();
    }

    public int getTotalPrepopulateCall() {
        return prepopulateCalls.size();
    }
}
